# export asset paths
from .base import DEFAULT_ASSETS, DEFAULT_SONG, DEFAULT_SOUND

# export functions and context managers
from .base import play_file, play_loop, play_while_running, play_after, \
  get_assets_dir, play_file_async, play_while_running_async, \
  play_after_async

# export bell
from .bell import bell, bell_after



from typing import Callable, Optional, Set
from multiprocessing import Process
from platform import platform
from pathlib import Path
from sys import exit
import signal


PLATFORM: str = platform().lower()
_PROCS: Set[Process] = set()


def play_process(
  file: Path,
  target: Callable,
  running_procs: Optional[Set[Process]] = _PROCS
) -> Process:
  proc = Process(target=target, args=(file,))

  if running_procs is not None:
    running_procs.add(proc)

  proc.start()

  return proc


def kill_process(
  proc: Process,
  running_procs: Optional[Set[Process]] = _PROCS
):
  proc.kill()
  proc.join()

  if running_procs is not None and proc in running_procs:
    running_procs.remove(proc)


def kill_child_procs(signum: int, frame):
  if _PROCS:
    for proc in _PROCS.copy():
      try:
        kill_process(proc)

      except:
        pass

  exit()


signal.signal(signal.SIGTERM, kill_child_procs)
signal.signal(signal.SIGINT, kill_child_procs)
signal.signal(signal.SIGSEGV, kill_child_procs)
signal.signal(signal.SIGABRT, kill_child_procs)
signal.signal(signal.SIGILL, kill_child_procs)

if 'windows' not in PLATFORM:
  signal.signal(signal.SIGPIPE, kill_child_procs)
  signal.signal(signal.SIGQUIT, kill_child_procs)
  signal.signal(signal.SIGBUS, kill_child_procs)



from typing import Callable, Any, Awaitable
from functools import wraps
from asyncio import Task, create_task

try:
  from asyncio import to_thread

except ImportError:
  from asyncio import get_running_loop

  async def to_thread(func: Callable, *args, **kwargs) -> Awaitable:
    loop = get_running_loop()

    return loop.run_in_executor(
      None,
      func,
      *args,
      **kwargs
    )


async def to_thread_task(func: Callable, *args, **kwargs) -> Task:
  coro = to_thread(func, *args, **kwargs)
  return create_task(coro)


async def finalize_task(task: Task) -> Any:
  try:
    return await task

  finally:
    if not task.cancelled():
      task.cancel()


CoroutineResult = Awaitable[Any]
CoroutineFunction = Callable[..., CoroutineResult]
CoroutineMethod = Callable[..., CoroutineResult]


def func_as_method_coro(func: Callable) -> CoroutineMethod:
  @wraps(func)
  async def method(self, *args, **kwargs) -> Any:
    return await to_thread(func, *args, **kwargs)

  return method


def coro_as_method_coro(coro: CoroutineFunction) -> CoroutineMethod:
  @wraps(coro)
  async def method(self, *args, **kwargs) -> Any:
    return await coro(*args, **kwargs)

  return method



from contextlib import contextmanager
from typing import ContextManager


BELL_CHAR = '\a'
NO_NEWLINE = ''


def bell():
  print(BELL_CHAR, end=NO_NEWLINE)


@contextmanager
def bell_after(play_bell: bool = True) -> ContextManager:
  try:
    yield

  finally:
    if play_bell:
      bell()



__author__ = "Alex DeLorenzo"
from setuptools import setup
from pathlib import Path


NAME = "play_sounds"
VERSION = "0.3.3"
LICENSE = "AGPL-3.0"

DESC = "🔊 Play music and sounds in your Python scripts"

REQUIREMENTS = \
  Path('requirements.txt') \
    .read_text() \
    .split('\n')

README = Path('README.md').read_text()


setup(
  name=NAME,
  version=VERSION,
  description=DESC,
  long_description=README,
  long_description_content_type="text/markdown",
  url="https://alexdelorenzo.dev",
  author=__author__,
  license=LICENSE,
  packages=[NAME],
  zip_safe=False,
  install_requires=REQUIREMENTS,
  python_requires='>=3.6',
  include_package_data=True,
  package_data={'play_sounds': ['assets/*']},
)



from contextlib import contextmanager, asynccontextmanager, \
    AbstractContextManager, AbstractAsyncContextManager
from typing import Callable, AsyncContextManager, Any, \
    ContextManager, Awaitable, Optional, Set
from multiprocessing import Process
from platform import platform
from pathlib import Path
from asyncio import sleep
import logging
import sys

from .wrap import to_thread
from .proc import play_process, kill_process


BLOCK_WHILE_PLAYING = True
PROCS = 1
DEFAULT_WAIT: float = 0.25


def get_assets_dir() -> Path:
  filename = __loader__.get_filename()
  return Path(filename).parent / 'assets'


DEFAULT_ASSETS = get_assets_dir()
DEFAULT_SONG = DEFAULT_ASSETS / 'song.mp3'
DEFAULT_SOUND = DEFAULT_ASSETS / 'ding.mp3'

PLATFORM = platform().lower()
MAJOR, MINOR, *_ = sys.version_info


if 'windows' in PLATFORM or 'nt' in PLATFORM:
  from playsound import playsound

  def play_file(file: Path, block: bool = BLOCK_WHILE_PLAYING):
    filename = str(file.absolute())
    playsound(filename, block=block)

    if not block:
      logging.warning("Playback must block on Windows.")

else:
  from boombox import BoomBox

  def play_file(file: Path, block: bool = BLOCK_WHILE_PLAYING):
    player = BoomBox(file, wait=block)
    player.play()


def play_loop(file: Path):
  try:
    while True:
      play_file(file)

  except Exception as e:
    logging.error(f"Error while trying to play {file}: {e}")


@contextmanager
def play_while_running(
  file: Path,
  loop: bool = True
) -> ContextManager[Process]:
  play_func = play_loop if loop else play_file
  proc = play_process(file, target=play_func)

  try:
    yield proc

  finally:
    kill_process(proc)


@contextmanager
def play_after(
  file: Path,
  block: bool = BLOCK_WHILE_PLAYING
) -> ContextManager[Path]:
  try:
    yield file

  finally:
    play_file(file, block)


async def play_file_async(
  file: Path,
  block: bool = BLOCK_WHILE_PLAYING,
  interval: float = DEFAULT_WAIT
):
  try:
    proc = await to_thread(play_process, file, target=play_file)

    while proc.is_alive():
      await sleep(interval)

  finally:
    await to_thread(kill_process, proc)


@asynccontextmanager
async def play_while_running_async(
  file: Path,
  loop: bool = True
) -> AsyncContextManager[Process]:
  play_func = play_loop if loop else play_file

  try:
    proc = await to_thread(play_process, file, target=play_func)
    yield proc

  finally:
    await to_thread(kill_process, proc)


@asynccontextmanager
async def play_after_async(
  file: Path,
  block: bool = BLOCK_WHILE_PLAYING,
  interval: float = DEFAULT_WAIT,
) -> AsyncContextManager[Path]:
  try:
    yield file

  finally:
    await play_file_async(file, block, interval)


#@dataclass
#class play_sound(AbstractContextManager, AbstractAsyncContextManager):
  #def __call__(self, func: Callable) -> Callable:
    #if iscoroutinefunction(func):
      #@wraps(func)
      #async def new_coro(*args, **kwargs) -> Awaitable[Any]:
        #async with self as proc:
          #pass
    #with play_while_running(file) as proc:
      #return func

  #def __enter__(self) -> ContextManager[Process]:
    #with play_while_running(file) as proc:
      #return proc

  #def __exit__(self, *args):
    #pass

  #async def __aenter__(self) -> Awaitable[AsyncContextManager[Process]]:
    #with play_while_running_async(file) as proc:
      #return proc


  #async def __aexit__(self, *args):
    #pass
